<?php

namespace Drupal\instagram_feeds;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\StateInterface;
use Drupal\instagram_feeds\Entity\InstagramAccountInterface;
use GuzzleHttp\ClientInterface;

/**
 * Instagram Feeds Cron Handler Service.
 *
 * Triggered inside hook_cron to import Instagram Posts.
 */
class CronHandler implements InstagramFeedsCronHandlerInterface {

  const SETTINGS = 'instagram_feeds.settings';
  const STATE_LAST_IMPORT_TIMESTAMP_PREFIX = 'instagram_feeds_last_imported_time_';

  /**
   * Avatar folder path.
   *
   * @var string
   */
  protected $avatarDir;

  /**
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The state storage service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * GuzzleHttp\ClientInterface definition.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Instance of Media storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $mediaStorage;

  /**
   * The Instagram Feeds Logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Entity Type Manager Service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The list of published active accounts.
   *
   * @var \Drupal\instagram_feeds\Entity\InstagramAccountInterface[]
   */
  protected $instagramAccounts;

  /**
   * The CronHandler service constructor.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    StateInterface $state,
    ClientInterface $http_client,
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelFactoryInterface $logger_factory
  ) {
    $config = $config_factory->get(self::SETTINGS);
    $this->config = $config;
    $this->entityTypeManager = $entity_type_manager;
    $this->state = $state;
    $this->httpClient = $http_client;
    $this->mediaStorage = $entity_type_manager->getStorage('media');
    $this->logger = $logger_factory->get('instagram_feeds');
    $this->avatarDir = $config->get('avatar_uri_scheme') . '://' . $config->get('avatar_directory');
  }

  /**
   * Gets active Instagram account entities.
   *
   * @return \Drupal\instagram_feeds\Entity\InstagramAccountInterface[]
   */
  protected function getInstagramAccounts() : array {
    if (!isset($this->instagramAccounts)) {
      $this->instagramAccounts = [];
      $instagram_accounts = $this->entityTypeManager->getStorage('instagram_account')->loadMultiple();
      /** @var \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account */
      foreach ($instagram_accounts as $account) {
        // Skip if entity is unpublished or token is invalid.
        if ($account->isPublished() && $account->tokenIsValid()) {
          $this->instagramAccounts[$account->id()] = $account;
        }
      }
    }
    return $this->instagramAccounts;
  }

  /**
   * {@inheritdoc}
   */
  public function importInstagramPosts() {
    foreach ($this->getInstagramAccounts() as $account) {
      $this->processAccount($account);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getLastImportTimestamp(InstagramAccountInterface $account): int {
    return $this->state->get(self::STATE_LAST_IMPORT_TIMESTAMP_PREFIX . $account->id(), 0);
  }

  /**
   * Gets 25 recent posts created by Instagram user.
   *
   * @param \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account
   *   Instagram account.
   *
   * @return array
   *   The list of Instagram posts starting from oldest to newest.
   *
   * @throws \Exception
   */
  protected function getMedia(InstagramAccountInterface $account) {
    try {
      $request_url = $account::INSTAGRAM_GRAPH_ENDPOINT . '/me/media?' . http_build_query([
        'fields' => 'id,caption,media_type,media_url,permalink,thumbnail_url,timestamp,username',
        'access_token' => $account->getToken(),
      ]);
      $body = $this->getInstagramResponceContents($request_url, TRUE);
      $result = array_filter($body['data'] ?: [], [$this, 'filterPostByPermalink']);
      $avatar = !$result ? [] : $this->getAvatarFromPostPermalink($account, $result[0]['permalink']);
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }

    foreach ($result as &$media) {
      $media['avatar'] = $avatar;
      $media['timestamp'] = strtotime($media['timestamp']);
      $media['full_name'] = $avatar['alt'] ?? $account->label();
      // If there is no caption, this field is missed in response.
      $media['caption'] = $media['caption'] ?: '';
      $media['tags'] = $this->parseHashTags($media['caption']);
    }
    return array_reverse($result);
  }

  /**
   * Filter callback to exclude posts without permalink.
   *
   * @param array $post
   *   Instagram Post array.
   *
   * @return bool
   *   TRUE if permalink is present in the post, FALSE otherwise.
   */
  protected function filterPostByPermalink($post) : bool {
    return (bool) $post['permalink'];
  }

  /**
   * Sends request to Instagram and validates response status code.
   *
   * @param string $url
   *   Url to get contents for.
   * @param $is_json_expected
   *   Set true, in order to decode JSON data from response.
   *
   * @return string|array
   *   Response body as string if $is_json_expected is false, or JSON decoded body.
   *
   * @throws \Exception
   */
  protected function getInstagramResponceContents($url, $is_json_expected = FALSE) {
    try {
      /** @var \Psr\Http\Message\ResponseInterface $response */
      $response = $this->httpClient->get($url);
      if ($response->getStatusCode() !== 200) {
        throw new \Exception(t('Invalid responce code @code from Instagram.', [
          '@code' => $response->getStatusCode(),
        ]));
      }
      $body = $response->getBody()->getContents();
      return $is_json_expected ? Json::decode($body) : $body;
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }
    return '';
  }

  /**
   * Scrape the Instagram post owner avatar URL using permalink URL.
   *
   * @param \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account
   *   Instagram account.
   * @param string $permalink_url
   *   Instagram post permalink URL.
   *
   * @return string[]
   *   An array with 'target_id' and 'alt' keys to fill in file reference field.
   *
   * @throws \Exception
   */
  protected function getAvatarFromPostPermalink(InstagramAccountInterface $account, $permalink_url) {
    try {
      $shards = explode('window._sharedData = ', $this->getInstagramResponceContents($permalink_url));
      $insta_encoded_json = explode(';</script>', $shards[1]);
      $insta_array = Json::decode($insta_encoded_json[0]);
      $owner = $insta_array['entry_data']['PostPage'][0]['graphql']['shortcode_media']['owner'] ?? [];

      if (!empty($owner['profile_pic_url']) && file_prepare_directory($this->avatarDir, FILE_CREATE_DIRECTORY)) {
        $file_data = file_get_contents($owner['profile_pic_url']);
        $full_name = $owner['full_name'] ?? 'Instagram Avatar';
        $avatar_file_extension = pathinfo(parse_url($owner['profile_pic_url'], PHP_URL_PATH), PATHINFO_EXTENSION);
        $file_uri_destination = $this->avatarDir . '/' . $account->id() . '.' . $avatar_file_extension;
        $file = file_save_data($file_data, $file_uri_destination, FILE_EXISTS_REPLACE);
      }

    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }

    return $file ? ['target_id' => $file->id(), 'alt' => $full_name] : [];
  }

  /**
   * Instagram posts import processor for the given account.
   *
   * @param \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account
   *   Instagram Feed configuration Entity.
   */
  protected function processAccount(InstagramAccountInterface $account) {
    $post_count = 0;
    $max_posts = $account->cron_import_limit->value;
    $user_last_imported_date = $this->getLastImportTimestamp($account);
    foreach ($this->getMedia($account) as $post) {
      // Stop foreach if we have already reached the limit.
      if ($post_count >= $max_posts) {
        break;
      }
      // Only save the insta post if its timestamp is after the saved last
      // import date to prevent duplicates.
      if ($post['timestamp'] > $user_last_imported_date) {
        $this->createMediaEntity($post, $account);
        $last_imported_timestamp = $post['timestamp'];
        $post_count++;
      }
    }
    // Update the last imported date only if there were new Instagram posts imported.
    if ($post_count > 0) {
      $this->state->set(self::STATE_LAST_IMPORT_TIMESTAMP_PREFIX . $account->id(), $last_imported_timestamp);
    }
    $logger_context = ['@account' => $account->label(), '@count' => $post_count];
    $this->logger->info("@count post(s) imported from @account account.<br />\n", $logger_context);
  }

  /**
   * Creates Media entity with Instagram data.
   *
   * @param array $post
   *   Instagram post data array.
   * @param \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account
   *   Instagram Feed configuration Entity.
   * @param \Drupal\file\FileInterface|bool $avatar_file
   *   Saved avatar image file.
   *
   * @todo Fields mapping between Instagram data and Media Entity Type.
   */
  protected function createMediaEntity(array $post, InstagramAccountInterface $account) {
    $media_type = $account->media_bundle->value;
    $permalink_field = $this->entityTypeManager->getStorage('media_type')->load($media_type)->getPluginCollections();
    $entity_array = ['bundle' => $media_type];
    // TODO: Make name configurable with tokens allowed.
    $entity_array['name'] = $account->label() . ' (' . date('m/d/Y', $post['timestamp']) . ') ' . implode(',', $post['tags']);
    $entity_array[$permalink_field['source_configuration']->getConfiguration()['source_field']] = $post['permalink'];

    $mapping = $this->config->get('mapping.' . $media_type);
    foreach ($mapping as $entity_field_name => $instagram_field_key) {
      if (!empty($instagram_field_key) && $post[$instagram_field_key]) {
        $entity_array[$entity_field_name] = $post[$instagram_field_key];
      }
    }

    $media_entity = $this->mediaStorage->create($entity_array);
    $media_entity->save();
  }

  /**
   * Parses hash tags from string.
   *
   * @param string $caption
   *
   * @return string[]
   *   An array of Instagram hash tags.
   */
  protected function parseHashTags($caption): array {
    $tags = [];
    preg_match_all('~(#\w+)~', $caption, $tags, PREG_PATTERN_ORDER);
    return $tags[1] ?: [];
  }

  /**
   * {@inheritdoc}.
   */
  public function refreshTokens() {
    $current_time = \Drupal::time()->getRequestTime();
    $frequency = $this->config->get('refresh_frequency');

    /** @var \Drupal\instagram_feeds\Entity\InstagramAccountInterface $account */
    foreach ($this->getInstagramAccounts() as $account) {
      // Long-lived tokens are valid 60 days, so token_expiration - 60 days
      // (5184000 sec) will be the date, when token was generated/refreshed
      // last time. Continue only if frequency period has gone.
      // Expired token can no longer be regenerated.
      if ($current_time > $account->token_expiration->value - 5184000 + $frequency) {
        $account->refreshToken($this->httpClient, TRUE);
      }
    }
  }

}
