<?php

namespace Drupal\instagram_feeds\Entity;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\Core\Url;
use GuzzleHttp\ClientInterface;

/**
 * InstagramAccount Interface.
 */
interface InstagramAccountInterface extends ContentEntityInterface, EntityPublishedInterface {

  const INSTAGRAM_GRAPH_ENDPOINT = 'https://graph.instagram.com';
  const INSTAGRAM_API_ENDPOINT = 'https://api.instagram.com';

  /**
   * Gets the Instagram long-lived token.
   *
   * @param string $client_id
   *   Instagram app (client) ID.
   * @param string $client_secret
   *   Instagram app (client) secret.
   * @param string $code
   *   Instagram auth code to regenerate the token.
   * @param bool $save
   *   Save entity or not after token has been refreshed successfully.
   *
   * @return string
   *   Instagram long-lived access token for the authorized Instagram account.
   */
  public function getToken($client_id = NULL, $client_secret = NULL, $code = NULL, $save = FALSE): string;

  /**
   * Tries to refresh long-lived Instagram access token.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   *   Guzzle HTTP Client.
   * @param bool $save
   *   Save entity or not after token has been refreshed successfully.
   *
   * @return bool
   *   True if success, false otherwise.
   *
   * @throws \Exception
   */
  public function refreshToken(ClientInterface $http_client, bool $save = FALSE): bool;

  /**
   * Checks if Instagram access token exists and hasn't expired.
   *
   * @return bool
   *   TRUE if token is set and has not been expired, FALSE otherwise.
   */
  public function tokenIsValid(): bool;

  /**
   * Gets URL to Instagram Auth form.
   *
   * @return \Drupal\Core\Url
   *   Url to Instagram Auth form.
   */
  public function getOauthUrl($client_id): Url;

  /**
   * Gets Instagram account username.
   *
   * @param bool $save
   *   Save entity or not when username obtained from Instagram API.
   *
   * @return string
   *   Instagram username.
   */
  public function getAccountName(bool $save = FALSE): string;

}
