<?php

namespace Drupal\instagram_feeds\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Provides the InstagramAccount add/edit form.
 */
class InstagramAccountForm extends ContentEntityForm {

  /**
   * Instagram App (Client) ID.
   *
   * @var string
   */
  protected $clientId;

  /**
   * Instagram App (Client) secret.
   *
   * @var string
   */
  protected $clientSecret;

  /**
   * Constructs a new instance.
   *
   * @param \Drupal\Core\Entity\EntityManagerInterface $entity_manager
   *   The entity manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface
   *   The configuration factory service.
   */
  public function __construct(
    EntityManagerInterface $entity_manager,
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory
  ) {
    parent::__construct($entity_manager);
    $this->entityTypeManager = $entity_type_manager;
    $config = $config_factory->get('instagram_feeds.settings');
    $this->clientId = $config->get('client_id');
    $this->clientSecret = $config->get('client_secret');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.manager'),
      $container->get('entity_type.manager'),
      $container->get('config.factory'),
      $container->get('date.formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    if (empty($this->clientId) || empty($this->clientSecret)) {
      return [
        'configure' => [
          '#type' => 'link',
          '#title' => $this->t('Configure Instagram Feeds'),
          '#access' => $this->currentUser()->hasPermission('administer instagram_feeds'),
          '#attributes' => [
            'class' => ['button', 'button--primary'],
          ],
          '#url' => Url::fromRoute('instagram_feeds.settings')->setOption('query', \Drupal::destination()->getAsArray()),
        ],
      ];
    }

    $i_account = $this->getEntity();

//    if ($i_account->isNew() && $i_account->get('token')->isEmpty()) {
//      return $this->getAuthWindowButton();
//    }
//    elseif (!$i_account->tokenIsValid()) {
    if (!$i_account->tokenIsValid()) {
      return $this->getAuthWindowButton() + parent::actions($form, $form_state);
    }

    return parent::actions($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    if (empty($this->clientId) || empty($this->clientSecret)) {
      return $form;
    }
    /** @var \Drupal\instagram_feeds\Entity\InstagramAccountInterface $i_account */
    $i_account = $this->getEntity();
    if ($this->operation == 'edit') {
      $form['#title'] = $this->t('<em>Edit @type</em> @title', [
        '@type' => $i_account->getEntityType()->getLabel(),
        '@title' => $i_account->label()
      ]);
    }
    $form = parent::form($form, $form_state);
    // TODO: Disable / hide fields with data from Instagram.
    $form['status']['#group'] = 'footer';

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function prepareEntity() {
    parent::prepareEntity();

    /** @var \Drupal\instagram_feeds\Entity\InstagramAccountInterface $i_account */
    $i_account = $this->getEntity();
    // Obtain Instagram short-lived and exchange to long-lived token.
    $code = $this->getRequest()->query->get('code');
    if ($i_account->isNew() && isset($code) && !empty($code)) {
      if ($i_account->get('token')->isEmpty() || $i_account->get('token_expiration')->isEmpty()) {
        $i_account->getToken($this->clientId, $this->clientSecret, $code);
      }
      if ($i_account->get('account')->isEmpty() && $i_account->tokenIsValid()) {
        $i_account->getAccountName();
      }
    }
  }

  /**
   * Gets link to Instagram Auth Window.
   *
   * @return array
   *   Form link element renderable array.
   */
  protected function getAuthWindowButton(): array {
    return [
      'auth' => [
        '#type' => 'link',
        '#title' => $this->t('Authenticate Instagram App'),
        '#attributes' => [
          'class' => ['button', 'button--primary'],
        ],
        '#url' => $this->getEntity()->getOauthUrl($this->clientId),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $instagram_account = $this->entity;
    $status = $instagram_account->save();
    if ($status === SAVED_NEW) {
      $this->messenger()->addMessage($this->t('The %label Instagram Account created.', [
        '%label' => $instagram_account->label(),
      ]));
    }
    else {
      $this->messenger()->addMessage($this->t('The %label Instagram Account updated.', [
        '%label' => $instagram_account->label(),
      ]));
    }
    $form_state->setRedirect('entity.instagram_account.collection');
    return $status;
  }

  /**
   * Helper function to check whether an Instagram Feed configuration exists.
   *
   * @param string $id
   *   Instagram Account ID.
   *
   * @return bool
   *   TRUE if Instagram Feed exists, FALSE otherwise.
   */
  public function instagramAccountIdExists($id) {
    $entity = $this->entityTypeManager->getStorage('instagram_account')->getQuery()
      ->condition('iid', $id)
      ->execute();
    return (bool) $entity;
  }

}
